#include "computescorethread.h"
#include <limits>
#include <cmath>
#include <iostream>
using namespace std;

ComputeScoreThread::ComputeScoreThread(QObject *parent)
    : QThread(parent)
{
    mRestart = false;
    mAbort = false;
	mNormalizedScoreMin = std::numeric_limits<double>::max();
	mNormalizedScoreMax = -std::numeric_limits<double>::max();
}

ComputeScoreThread::~ComputeScoreThread()
{
    mMutex.lock();
    mAbort = true;
    mCondition.wakeOne();
    mMutex.unlock();

    wait();
}

void ComputeScoreThread::Compute(MlfModelData::EInputScoreType inputScoreType, MlfModelData::EScoreType scoreType, double prior, double multiplicative, double additive, double threshold)
{
    QMutexLocker locker(&mMutex);

    mInputScoreType = inputScoreType;
    mScoreType = scoreType;
	mPrior = prior;
	mMultiplicative = multiplicative;
	mAdditive = additive;
	mThreshold = threshold;

    if (!isRunning()) {
        start(LowPriority);
    } else {
        mRestart = true;
        mCondition.wakeOne();
    }
}

void ComputeScoreThread::run()
{
    forever {
        mMutex.lock();
		MlfModelData::EScoreType scoreType = mScoreType;
		double prior = mPrior;
		double multiplicative = mMultiplicative;
		double additive = mAdditive;
		double threshold = mThreshold;
        mMutex.unlock();


		DBG("ComputeScoreThread start...");

		double normalizedScoreMin = std::numeric_limits<double>::max();
		double normalizedScoreMax = -std::numeric_limits<double>::max();

		for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
		{
			if (mRestart) {
				DBG("ComputeScoreThread restarted");
				break;
			}
			if (mAbort)
				return;

			MlfModelData::MlfTableRow *row = *i;
			double score1 = row->mpMlfRecord->mScore1;
			double score2 = 0;
	//		DBG("score1:"<<score1<<" exp:"<<exp(score1)<<" logit-exp:"<<logit(exp(score1)));
			double llr = 0;
			double posterior = 0;
			switch (mInputScoreType)
			{
				case MlfModelData::InputScoreTypeLogprob:
					if (score1 >= 0) 
					{
						llr = 15 + prior;
						posterior = 1;
					} else {
						posterior = exp(score1);
						llr = logit(posterior) + prior;
					}
					break;

				case MlfModelData::InputScoreTypeLlr:
					posterior = sigm(score1);
					llr = score1 + prior;
					break;
			}
			switch (scoreType)
			{
				case MlfModelData::LogLikelihoodRatio:
					score2 = llr;
					break;

				case MlfModelData::KarlovcoScore:
					//DBG("sigm("<<llr<<" * "<<multiplicative<<" + "<<additive<<")" );
					score2 = sigm((llr - additive) * multiplicative);
					break;
			}
			row->mScorePosterior = posterior;
			row->mScoreLlr = llr;
			row->mScoreLlrNorm = (llr - additive) * multiplicative;
			row->mpMlfRecord->mScore2 = score2;
			row->mpMlfRecord->mHardDecision = score2 > threshold;
			if (normalizedScoreMax < score2) { 
				normalizedScoreMax = score2; 
			}
			if (normalizedScoreMin > score2) { 
				normalizedScoreMin = score2; 
			}
		}

		if (!mRestart)
		{
			mpHistogramPlot->Reset(normalizedScoreMin, normalizedScoreMax);
			mpHistogramPlot->SetAdditiveConstant(additive);
			mpHistogramPlot->SetMultiplicativeConstant(multiplicative);
		}
		for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
		{
			if (mRestart || mAbort)
			{
				break;
			}
			if ((*mpSelectedKeywords)[(*i)->mLabelUnicode])
			{
				mpHistogramPlot->AddData(*i);
			}
		}
		mpHistogramPlot->PostProcessData(this);

		if (!mRestart)
		{
			mNormalizedScoreMin = normalizedScoreMin;
			mNormalizedScoreMax = normalizedScoreMax;
			emit ScoreComputed(normalizedScoreMin, normalizedScoreMax);
			mpHistogramPlot->SetThreshold(threshold);
			mpHistogramPlot->update();
		}

		DBG("ComputeScoreThread finished");

        mMutex.lock();
        if (!mRestart)
            mCondition.wait(&mMutex);
        mRestart = false;
        mMutex.unlock();
	}
}

double ComputeScoreThread::logit(double p)
{
	if (p <= 0) { return -std::numeric_limits<double>::max(); }
	if (p >= 1) { return std::numeric_limits<double>::max(); }
	return log(p/(1-p));
}

double ComputeScoreThread::sigm(double d)
{
	return 1/(1+exp(-d));
}

double ComputeScoreThread::Llr2Posterior(double llr, double prior)
{
	return sigm(llr) + prior;
}

double ComputeScoreThread::LogLik2Posterior(double logLik, double prior)
{
	return logLik >= 0 ? 1 : exp(logLik) + prior;
}

double ComputeScoreThread::Posterior2Llr(double posterior, double prior)
{
	return posterior >= 1 ? 15 + prior : logit(posterior) + prior;
}

