#include "mediaplayerthread.h"
#include <QtGui>

#define ERRCHECK(msg) \
    if (result != FMOD_OK) \
    { \
		emit Error(tr("FMOD error! (").append(result).append(") \"").append(FMOD_ErrorString(result)).append(tr("\""))); \
		return; \
    }

#define ERRCHECK_RETVAL(msg, retval) \
    if (result != FMOD_OK) \
    { \
		emit Error(tr("FMOD error! (").append(result).append(") \"").append(FMOD_ErrorString(result)).append(tr("\""))); \
		return retval; \
    }

#define ERRCHECK_POS(msg, pos) \
    if (result != FMOD_OK) \
    { \
		emit Error(tr("FMOD error! (").append(result).append(") \"").append(FMOD_ErrorString(result)).append(tr("\"")).append(tr(" ...you are trying to seek to position %1 ms but the length of the audio is only %2 ms. If you are playing a raw file, you should change sampling rate or format in Preferences.").arg(pos).arg(mAudioLengthMs))); \
		return; \
    }


MediaPlayerThread::MediaPlayerThread(QObject *parent)
    : QThread(parent), mSeek(false), mUpdateIntervalMs(100)
{
    mRestart = false;
    mAbort = false;
	mContextMs = false;
	mState = StateStopped;
	mLengthMsPrevious = 0;

    FMOD_RESULT       result;

	mpFmodSound = NULL;

    result = FMOD_System_Create(&mpFmodSystem);
    ERRCHECK(result);

    result = FMOD_System_GetVersion(mpFmodSystem, &mFmodVersion);
    ERRCHECK(result);

    if (mFmodVersion < FMOD_VERSION)
    {
		emit Error(tr("FMOD error! You are using an old version of FMOD ").append(mFmodVersion).append(". This program requires ").append(FMOD_VERSION));
        return;
    }

#ifndef WIN32
	result = FMOD_System_SetOutput(mpFmodSystem, FMOD_OUTPUTTYPE_ALSA);
	ERRCHECK(result);
#else
	result = FMOD_System_SetOutput(mpFmodSystem, FMOD_OUTPUTTYPE_AUTODETECT);
    ERRCHECK(result);
#endif

    result = FMOD_System_Init(mpFmodSystem, 1, FMOD_INIT_NORMAL, 0);
    ERRCHECK(result);

	mAudioFileSuffix = "wav";
	mIsRaw = false;
	memset(&mExinfo, 0, sizeof(FMOD_CREATESOUNDEXINFO));
	mExinfo.cbsize = sizeof(FMOD_CREATESOUNDEXINFO);
	mExinfo.numchannels = 1;
	mExinfo.defaultfrequency = 8000;
	mExinfo.format = FMOD_SOUND_FORMAT_PCM16;
}

MediaPlayerThread::~MediaPlayerThread()
{
	DBG("MediaPlayerThread::~MediaPlayerThread()");
    mMutex.lock();
    mAbort = true;
    mCondition.wakeOne();
    mMutex.unlock();

    wait();

	DBG("releasing FMOD");
    FMOD_RESULT       result;
	if (mpFmodSound)
	{
	    result = FMOD_Sound_Release(mpFmodSound);
    	ERRCHECK(result);
	}
    result = FMOD_System_Close(mpFmodSystem);
    ERRCHECK(result);
    result = FMOD_System_Release(mpFmodSystem);
    ERRCHECK(result);
	DBG("~MediaPlayerThread done");
}
/*
void MediaPlayerThread::Play(const QString& filename, qint64 seekMs, qint64 stopMs)
{
	SetFilename(filename);
	SetSeekMs(seekMs);
	SetStopMs(stopMs);
	PlayDetection();
}
*/
bool MediaPlayerThread::LoadAudioFile()
{
    FMOD_RESULT       result;

	QString audio_filename = mAudioFilesDirectory;
	audio_filename.append(QDir::separator()).append(mFilename).append(".").append(mAudioFileSuffix);
	if (!QFile::exists(audio_filename))
	{
		QString audio_filename_0 = audio_filename;
		DBG("mLabPath: "<<mLabPath.toStdString());
		audio_filename = mLabPath;
		audio_filename.append(QDir::separator()).append(mFilename).append(".").append(mAudioFileSuffix);

		DBG("exists("<<audio_filename.toStdString()<<")");
		if (!QFile::exists(audio_filename))
		{
			emit Error(tr("The audio file \"%1\" nor \"%2\" do not exist").arg(audio_filename_0, audio_filename));
			return false;
		}
	}

	if (mIsRaw)
	{
		QFile f(audio_filename);
		mExinfo.length = f.size();
    	result = FMOD_System_CreateSound(mpFmodSystem, audio_filename.toStdString().c_str(), FMOD_HARDWARE | FMOD_2D | FMOD_OPENRAW, &mExinfo, &mpFmodSound);
	}
	else
	{
	    result = FMOD_System_CreateStream(mpFmodSystem, audio_filename.toStdString().c_str(), FMOD_HARDWARE | FMOD_2D, 0, &mpFmodSound);
	}
    ERRCHECK_RETVAL(result, false);
	
	result = FMOD_Sound_GetLength(mpFmodSound, &mAudioLengthMs, FMOD_TIMEUNIT_MS);
	ERRCHECK_RETVAL(result, false);

//	mpMediaObject->setCurrentSource(audio_filename);
	mLoadedFilename = mFilename;
	return true;
}

void MediaPlayerThread::Play()
{
    QMutexLocker locker(&mMutex);

	FMOD_BOOL paused; 
	FMOD_Channel_GetPaused(mpFmodChannel, &paused); 
	if (paused)
	{
		FMOD_Channel_SetPaused(mpFmodChannel, false); 
		SetState(StatePlaying);
	}
	else
	{
		mPlayDetection = false;

		if (!isRunning()) {
			start(LowPriority);
		} else {
			mRestart = true;
			mCondition.wakeOne();
		}
	}
}

void MediaPlayerThread::PlayDetection()
{
    QMutexLocker locker(&mMutex);

	mPlayDetection = true;

    if (!isRunning()) {
        start(LowPriority);
    } else {
        mRestart = true;
        mCondition.wakeOne();
    }
}

void MediaPlayerThread::run()
{
    FMOD_RESULT       result;
    forever {
        mMutex.lock();
		// copy data to local variables
        mMutex.unlock();

		if (mFilename != mLoadedFilename)
		{
			if (!LoadAudioFile())
			{
				return; // error
			}
		}
		result = FMOD_System_PlaySound(mpFmodSystem, FMOD_CHANNEL_FREE, mpFmodSound, false, &mpFmodChannel);
		ERRCHECK(result);
		if (mPlayDetection)
		{
			DBG("PlayDetection: "<<mDetectionStartMs);
			result = FMOD_Channel_SetPosition(mpFmodChannel, mDetectionStartMs, FMOD_TIMEUNIT_MS);
			ERRCHECK_POS(result, mDetectionStartMs);
		}
		else if (mSeek)
		{
			result = FMOD_Channel_SetPosition(mpFmodChannel, mSeekMs, FMOD_TIMEUNIT_MS);
			ERRCHECK_POS(result, mSeekMs);
			mSeek = false;
		}

		while(1)
		{
			if (mRestart) {
				break;
			}
			if (mAbort)
				return;

			result = FMOD_System_Update(mpFmodSystem);
			ERRCHECK(result);

			if (mpFmodChannel)
			{
				FMOD_BOOL         playing;
				FMOD_BOOL         paused;

				if (mSeek)
				{
					result = FMOD_Channel_SetPosition(mpFmodChannel, mSeekMs, FMOD_TIMEUNIT_MS);
					ERRCHECK_POS(result, mSeekMs);
					mSeek = false;
				}

				result = FMOD_Channel_IsPlaying(mpFmodChannel, &playing);
				if ((result != FMOD_OK) && (result != FMOD_ERR_INVALID_HANDLE))
				{
					ERRCHECK(result);
				}
				if (playing)
				{
					SetState(StatePlaying);
				}
				else
				{
					result = FMOD_Channel_GetPaused(mpFmodChannel, &paused);
					if ((result != FMOD_OK) && (result != FMOD_ERR_INVALID_HANDLE))
					{
						ERRCHECK(result);
					}
					if (paused)
					{
						SetState(StatePaused);
					}
					else
					{
						SetState(StateStopped);
						break;
					}
				}

				result = FMOD_Channel_GetPosition(mpFmodChannel, &mPositionMs, FMOD_TIMEUNIT_MS);
				if ((result != FMOD_OK) && (result != FMOD_ERR_INVALID_HANDLE))
				{
					ERRCHECK(result);
				}
				if (mPlayDetection && mPositionMs >= mDetectionEndMs)
				{
					result = FMOD_Channel_Stop(mpFmodChannel);
					ERRCHECK(result);
					SetState(StateStopped);
				}
				emit AudioFilePositionChanged(mPositionMs);

				result = FMOD_Sound_GetLength(mpFmodSound, &mLengthMs, FMOD_TIMEUNIT_MS);
				if ((result != FMOD_OK) && (result != FMOD_ERR_INVALID_HANDLE))
				{
					ERRCHECK(result);
				}
				if (mLengthMs != mLengthMsPrevious)
				{
					emit AudioFileLengthChanged(mLengthMs);
					mLengthMsPrevious = mLengthMs;
				}
			}

			msleep(mUpdateIntervalMs);
		}
		if (!mRestart)
		{
			SetState(StateStopped);
		}

        mMutex.lock();
        if (!mRestart)
            mCondition.wait(&mMutex);
        mRestart = false;
        mMutex.unlock();
	}
}

void MediaPlayerThread::SetState(EState state)
{
	if (mState != state)
	{
		mState = state;
		switch(state)
		{
			case StatePlaying: emit PlayingStarted(); break;
			case StatePaused: emit Paused(); break;
			case StateStopped: emit Stopped(); break;
		}
	}
}

void MediaPlayerThread::Pause()
{
	FMOD_BOOL paused; 
	FMOD_Channel_GetPaused(mpFmodChannel, &paused); 
	FMOD_Channel_SetPaused(mpFmodChannel, !paused); 
	if (paused)
	{
		SetState(StatePlaying);
	}
	else
	{
		SetState(StatePaused);
	}
}

void MediaPlayerThread::Stop()
{
	if (mState == StatePlaying || mState == StatePaused)
	{
		FMOD_Channel_Stop(mpFmodChannel);
		SetState(StateStopped);
	}
}

/*
void MediaPlayerThread::ERRCHECK(FMOD_RESULT result)
{
    if (result != FMOD_OK)
    {
		emit Error(tr("FMOD error! (").append(result).append(") \"").append(FMOD_ErrorString(result)).append(tr("\"")));
    }
}
*/

void MediaPlayerThread::SeekTo(unsigned int posMs)
{
	mSeekMs = posMs;
	mSeek = true;
}

void MediaPlayerThread::SetDetectionStartMs(qint64 seekMs) 
{ 
	DBG("MediaPlayerThread::SetDetectionStartMs("<<seekMs<<") mContextMs:"<<mContextMs);
	mDetectionStartMs = seekMs; 
	if (mDetectionStartMs > mContextMs) { mDetectionStartMs -= mContextMs; } else { mDetectionStartMs = 0; }
	DBG("MediaPlayerThread::SetDetectionStartMs("<<mDetectionStartMs<<")");
}

void MediaPlayerThread::SetDetectionEndMs(qint64 stopMs) 
{ 
	DBG("MediaPlayerThread::SetDetectionEndMs("<<stopMs<<")");
	mDetectionEndMs = stopMs + mContextMs; 
}

