#include <QtGui>
#include <fstream>
#include <limits>
#include <sstream>
#include "mlfmodel.h"

using namespace std;

MlfModel::MlfModel(QObject *parent)
	: QAbstractTableModel(parent), mInputScoreType(MlfModelData::InputScoreTypeLogprob), mScoreType(MlfModelData::LogLikelihoodRatio), mThreshold(0), mPrior(0), mMultiplicative(1), mAdditive(0), mpHistogramFrame(NULL)
{
//	mHeaderData << "Filename" << "Start" << "End" << "Label" << "Raw score" << "Normalized score" << "HD" << "Correct";
	mpData = new MlfModelData::Data;
	mpSelectedKeywords = NULL;
	mComputeScoreThread.SetData(mpData);
	mReferenceBoundaryMs = 0;

	connect(&mComputeScoreThread, SIGNAL(ScoreComputed(double, double)), this, SLOT(NormalizedScoreRecomputed(double, double)));
}

MlfModel::~MlfModel()
{
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		delete *i;
	}
}

int MlfModel::rowCount(const QModelIndex & parent) const
{
	Q_UNUSED(parent);
	//DBG("rowCount()" << mpData->size() );
	return mpData->size();
}

int MlfModel::columnCount(const QModelIndex & parent) const
{
	Q_UNUSED(parent);
	//DBG("columnCount(): " << mHeaderData.size() );
	return MlfModelData::Columns::TotalCount();
}

QVariant MlfModel::data(const QModelIndex &index, int role) const
{
	//DBG("data() row:"<<index.row()<<" column:"<<index.column() );
    if (!index.isValid())
        return QVariant();

    if (role != Qt::DisplayRole/* && index.column() != 7*/)
        return QVariant();

	QVariant value = (*(*mpData)[index.row()])[index.column()];

	if ((index.column() == MlfModelData::EColStartTime || index.column() == MlfModelData::EColEndTime))
	{
		QString s;
		s.sprintf("%.2f", value.toDouble()/1000);
		return s;
	}
	else if ((index.column() == MlfModelData::EColScoreRaw || index.column() == MlfModelData::EColScorePosterior || index.column() == MlfModelData::EColScoreLlr || index.column() == MlfModelData::EColScoreLlrNorm || index.column() == MlfModelData::EColScoreNorm))
	{
		QString s;
		s.sprintf("%.6f", value.toDouble());
		return s;
/*		if (s == "0.0000") {
			return value;
		} else {
			return s;
		}
*/	}
	else
	{
		return value;
	}
}

QVariant MlfModel::headerData(int section,
		Qt::Orientation orientation,
		int role) const
{
	//DBG("MlfModel::headerData");
	if (role != Qt::DisplayRole)
		return QVariant();

	if (orientation == Qt::Horizontal)
		return MlfModelData::Columns::GetColumnLabel((MlfModelData::EColumn)section);
	else
		return QString("%1").arg(section+1);
//		return QVariant();
}

/*
void MlfModel::sort ( int column, Qt::SortOrder order)
{
	
}
*/
int MlfModel::SetItem( int row, const MlfModelData::MlfTableRow& data )
{
	//--- validate the index ---
	if ( row < 0 || row >= rowCount() )
		return 0;	// error!

	//--- sets the data ---
	*((*mpData)[row]) = data;

	//--- tell the world ---
	emit dataChanged( createIndex( row, 0 ), createIndex( row, columnCount() ) );

	return 1;		// success!
}

void MlfModel::LoadFromFile(const QString& filename)
{
	const int mlf_scanner_max_progress = 90;
	MlfScanner lexer;
	lexer.SetMlfScannerProcessingDisplay(mpProgressDisplay);
	lexer.SetProgressMaximum(mlf_scanner_max_progress);
	DBG("MlfModel::LoadFromFile() parsing...");
	mMlf.Clear();
	lexer.ParseFile(filename.toStdString(), &mMlf);
	DBG("MlfModel::LoadFromFile() parsing...done");
	//DBG("mMlf:" << mMlf );

//	beginRemoveRows( null, 0, rowCount()-1 );
	mpData->clear();
	mMlfLabels.clear();
	DBG("mpData->size():"<< mpData->size());
//	endRemoveRows();

	DBG("codecForName("<<mCharsetName.toStdString()<<")");
	QTextCodec *codec = QTextCodec::codecForName(mCharsetName.toStdString().c_str());

//	beginInsertRows( NULL, 0, mMlf.GetRecordsCount()-1 );
//	beginInsertColumns( NULL, 0, 5 );
	mNormalizedScoreMin = numeric_limits<double>::max();
	mNormalizedScoreMax = -numeric_limits<double>::max();
	unsigned int i = 0;
	for (Mlf::iterator iFile=mMlf.begin(); iFile!=mMlf.end(); iFile++, i++)
	{
		mpProgressDisplay->MlfScannerProgress(mlf_scanner_max_progress + i/mMlf.size()*50);
		for (MlfOneFile::iterator iRec=iFile->second->begin(); iRec!=iFile->second->end(); iRec++)
		{
			MlfRecord* pRec = *iRec;
			MlfModelData::MlfTableRow* row = new MlfModelData::MlfTableRow;

			row->mpMlfRecord = pRec;
			row->mFilename = ((QString)iFile->second->GetFilename().c_str()).remove(QRegExp("^.*[/\\\\]")).remove("\"").remove(QRegExp("\\.[^\\.]*$")).toStdString();
			row->mLabPath = ((QString)iFile->second->GetFilename().c_str()).remove(QRegExp("[^/\\\\]*$")).remove("\"");
			//DBG("filename:"<<iFile->second->GetFilename()<<" mLabPath:"<<row->mLabPath.toStdString());
			row->mLabelUnicode = codec->toUnicode(pRec->mLabel.c_str());
			//row->mFilename = iFile->second->GetFilename();
			if (mInputScoreType == MlfModelData::InputScoreTypeLogprob)
			{
				row->mScorePosterior = ComputeScoreThread::LogLik2Posterior(row->mpMlfRecord->mScore1);
				row->mScoreLlr = ComputeScoreThread::Posterior2Llr(row->mScorePosterior);
				row->mScoreLlrNorm = ComputeScoreThread::Posterior2Llr(row->mScorePosterior);
			}
			else if (mInputScoreType == MlfModelData::InputScoreTypeLlr)
			{
				row->mScorePosterior = ComputeScoreThread::Llr2Posterior(row->mpMlfRecord->mScore1);
				row->mScoreLlr = row->mpMlfRecord->mScore1;
				row->mScoreLlrNorm = row->mpMlfRecord->mScore1;
			}
			//DBG("posterior:"<<row->mScorePosterior<<" llr:"<<row->mScoreLlr);
			
			MlfLabels::iterator iLab = mMlfLabels.find((QString)row->mpMlfRecord->mLabel.c_str());
			if (iLab == mMlfLabels.end())
			{
				mMlfLabels.insert((QString)row->mpMlfRecord->mLabel.c_str(), 1);
			}
			else
			{
				iLab.value()++;
			}

			if (mNormalizedScoreMax < pRec->mScore2) { 
				mNormalizedScoreMax = pRec->mScore2; 
			}
			if (mNormalizedScoreMin > pRec->mScore2) { 
				mNormalizedScoreMin = pRec->mScore2; 
			}
			//DBG("row: " << *row );
			mpData->push_back(row);
		}
	}
	DBG("mMlfLabels.size(): "<<mMlfLabels.size());
}

void MlfModel::LoadRefFromFile(const QString& filename)
{
	const int mlf_scanner_progress_length = 50;
	const int mlf_paths_progress_length = 50;
//	const int mlf_compare_progress_length = 40;
	MlfScanner lexer;
	Mlf ref_mlf;
	lexer.SetMlfScannerProcessingDisplay(mpProgressDisplay);
	lexer.SetProgressMaximum(mlf_scanner_progress_length);
	DBG("MlfModel::LoadRefFromFile() parsing...");
	lexer.ParseFile(filename.toStdString(), &ref_mlf);
	DBG("MlfModel::LoadRefFromFile() parsing...done");
	//DBG("mRefMlf:" << mRefMlf );

	DBG("MlfModel::LoadRefFromFile() paths...");
	mRefMlf.Clear();
	// remove path and filename extension
	unsigned int i = 0;
	for (Mlf::iterator iFile=ref_mlf.begin(); iFile!=ref_mlf.end(); iFile++, i++)
	{
		std::string s = ((QString)iFile->second->GetFilename().c_str()).remove(QRegExp("^.*[/\\\\]")).remove("\"").remove(QRegExp("\\.[^\\.]*$")).toStdString();
		iFile->second->SetFilename(s);
		mRefMlf.insert(make_pair(s, iFile->second));
		mpProgressDisplay->MlfScannerProgress(mlf_scanner_progress_length + i/ref_mlf.size()*mlf_paths_progress_length);
//		DBG("ref filename: "<<iFile->second->GetFilename() );
	}
	RefreshReference();
	DBG("MlfModel::LoadRefFromFile() paths...done");

	DBG("MlfModel::LoadRefFromFile() filling rows...");
}

void MlfModel::RefreshReference()
{
	for (Mlf::iterator iFile = mRefMlf.begin(); iFile != mRefMlf.end(); iFile++)
	{
		for (MlfOneFile::iterator iRefRec=iFile->second->begin(); iRefRec!=iFile->second->end(); iRefRec++)
		{
			MlfRecord* pRefRec = *iRefRec;
			pRefRec->mHardDecision = 0;
		}
	}

	unsigned int i = 0;
	for (MlfModelData::Data::iterator iData=mpData->begin(); iData!=mpData->end(); iData++, i++)
	{
		MlfModelData::MlfTableRow* row = *iData;
		row->mpRefMlfRecord = NULL;
		row->mCorrect = MlfModelData::CorrectNo;
		// find in the reference
//		DBG("row: "<<*row );
		Mlf::iterator iFile = mRefMlf.find(row->mFilename);
		if (iFile != mRefMlf.end())
		{
//			DBG("iFile found" );
			for (MlfOneFile::iterator iRefRec=iFile->second->begin(); iRefRec!=iFile->second->end(); iRefRec++)
			{
				MlfRecord* pRefRec = *iRefRec;
//				DBG("pRefRec: "<<*pRefRec);
				// if labels are the same
				if (row->mpMlfRecord->mLabel == pRefRec->mLabel)
				{
//					DBG("labels match" );
					unsigned long mid = row->mpMlfRecord->mStartTime + (unsigned long)((row->mpMlfRecord->mEndTime - row->mpMlfRecord->mStartTime)/2);
					// if the hypothesis is correct
					unsigned long boundary_left = pRefRec->mStartTime > mReferenceBoundaryMs ? pRefRec->mStartTime - mReferenceBoundaryMs : 0;
					unsigned long boundary_right = pRefRec->mEndTime + mReferenceBoundaryMs;
					DBG("mStartTime:"<<pRefRec->mStartTime<<" - "<<mReferenceBoundaryMs<<" < "<<mid<<" && "<<mid<<" < "<<pRefRec->mEndTime<<" + "<<mReferenceBoundaryMs);
					if (pRefRec->mHardDecision == 0 && boundary_left < mid && mid < boundary_right)
					{
						DBG("overlapping" );
						row->mpRefMlfRecord = pRefRec;
						row->mCorrect = MlfModelData::CorrectYes;
						pRefRec->mHardDecision++;
					}
				}
			}
		}
		else
		{
			row->mCorrect = MlfModelData::CorrectUndefined;
		}
		mpProgressDisplay->MlfScannerProgress(i/mpData->size()*100);
	}

	unsigned int reference_misses = 0;
	unsigned int references = 0;
	for (Mlf::iterator iFile = mRefMlf.begin(); iFile != mRefMlf.end(); iFile++)
	{
		for (MlfOneFile::iterator iRefRec=iFile->second->begin(); iRefRec!=iFile->second->end(); iRefRec++)
		{
			MlfRecord* pRefRec = *iRefRec;
			MlfLabels::iterator i = mMlfLabels.find((QString)pRefRec->mLabel.c_str());
			// if the label should be scored
			if (i != mMlfLabels.end())
			{
				references++;
				if (pRefRec->mHardDecision == 0)
				{
					reference_misses++;
				}
				else if (pRefRec->mHardDecision > 1)
				{
					CERR("more than 1 detection for reference: "<<*pRefRec);
					EXIT();
				}
			}
		}
	}
	DBG("reference_misses:"<<reference_misses);
	DBG("references:"<<references);
	mComputeScoreThread.GetHistogramPlot()->SetReferenceMisses(reference_misses);
	DBG("GetReferenceMisses: "<<mComputeScoreThread.GetHistogramPlot()->GetReferenceMisses());

	DBG("MlfModel::LoadRefFromFile() filling rows...done");
	RefreshHistogram();
	emit dataChanged( createIndex( 0, MlfModelData::EColCorrect ), createIndex( rowCount()-1, MlfModelData::EColCorrect ) );
}

void MlfModel::RefreshHistogram()
{
	HistogramPlot * histogram_plot = mComputeScoreThread.GetHistogramPlot();
	histogram_plot->Reset(mNormalizedScoreMin, mNormalizedScoreMax);
	for (MlfModelData::Data::iterator iData=mpData->begin(); iData!=mpData->end(); iData++)
	{
		MlfModelData::MlfTableRow* row = *iData;
		if ((*mpSelectedKeywords)[row->mLabelUnicode])
		{
			histogram_plot->AddData(row);
		}
	}
	histogram_plot->PostProcessData();
}

void MlfModel::SaveMlfToFile(const Mlf& mlf, const QString& filename)
{
	ofstream f(filename.toStdString().c_str());
	if (!f.good())
	{
		QMessageBox::critical(NULL, tr("MLF Toy"), tr("Can not write to the output file"), QMessageBox::Ok, QMessageBox::Ok);
		return;
	}

	QString header = " norm_multiplicative=%1 norm_subtractive=%2";
	header.arg(mMultiplicative).arg(mAdditive);
	f << "#!MLF!# " << header.toStdString() << endl;

	for (Mlf::const_iterator iFile=mlf.begin(); iFile!=mlf.end(); iFile++)
	{
		f << "\"" << iFile->second->GetFilename() << "\"" << endl;
		for (MlfOneFile::const_iterator iRec=iFile->second->begin(); iRec!=iFile->second->end(); iRec++)
		{
			const MlfRecord* pRec = *iRec;
			f 
				<< pRec->mStartTime << "0000 " 
				<< pRec->mEndTime << "0000 " 
				<< pRec->mLabel << " " 
				<< setprecision(10)
				<< pRec->mScore1 << " " 
				<< pRec->mScore2 << " " 
				<< pRec->mHardDecision
				<< endl;
		}
		f << "." << endl;
	}
	f.close();
}

void MlfModel::SaveToFile(const QString& filename)
{
	SaveMlfToFile(mMlf, filename);
}

void MlfModel::SaveRefToFile(const QString& filename)
{
	ofstream f(filename.toStdString().c_str());
	if (!f.good())
	{
		QMessageBox::critical(NULL, tr("MLF Toy"), tr("Can not write to the output file"), QMessageBox::Ok, QMessageBox::Ok);
		return;
	}
	Mlf mlf;
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		MlfModelData::MlfTableRow * row = *i;
		if (row->mCorrect == MlfModelData::CorrectYes)
		{
			std::ostringstream rec_filename;
			//rec_filename << row->mLabPath.toStdString() << "/" << row->mFilename << ".rec";
			rec_filename << "*/" << row->mFilename << ".lab";
			MlfOneFile * mlf_one_file = NULL;
			Mlf::iterator i_file;
			if ((i_file = mlf.find(rec_filename.str())) == mlf.end())
			{
				mlf_one_file = new MlfOneFile;
				mlf_one_file->SetFilename(rec_filename.str());
				mlf.Add(mlf_one_file);
			}
			else
			{
				mlf_one_file = i_file->second;
			}

			if (row->mpRefMlfRecord)
			{
				mlf_one_file->Add(row->mpRefMlfRecord);
			}
			else
			{
				mlf_one_file->Add(row->mpMlfRecord);
			}
		}
	}

	SaveMlfToFile(mlf, filename);
}

void MlfModel::SetEncoding(QString charsetName)
{
	mCharsetName = charsetName;
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		MlfModelData::MlfTableRow *row = *i;
		QTextCodec *codec = QTextCodec::codecForName(charsetName.toStdString().c_str());
		row->mLabelUnicode = codec->toUnicode(row->mpMlfRecord->mLabel.c_str());
	}
	if (rowCount() > 0)
	{
		emit dataChanged( createIndex( 0, MlfModelData::EColLabel ), createIndex( rowCount()-1, MlfModelData::EColLabel ) );
	}
}

void MlfModel::SetInputScoreType(MlfModelData::EInputScoreType inputScoreType)
{
	DBG("MlfModel::SetInputScoreType(): "<<inputScoreType);
	mInputScoreType = inputScoreType;
	RecomputeNormalizedScore();
/*	if (rowCount() > 0)
	{
		emit dataChanged( createIndex( 0, MlfModelData::EColScorePosterior ), createIndex( rowCount()-1, MlfModelData::EColHardDecision ) );
	}
*/	
}

void MlfModel::ThresholdChanged(double t)
{	
	mThreshold = t;
//	RecomputeNormalizedScore();
	
	HistogramPlot * histogram_plot = mComputeScoreThread.GetHistogramPlot();
//	histogram_plot->Reset(mNormalizedScoreMin, mNormalizedScoreMax);
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		MlfModelData::MlfTableRow *row = *i;
		row->mpMlfRecord->mHardDecision = row->mpMlfRecord->mScore2 > t ? 1 : 0;
//		histogram_plot->AddData(row);
	}
	histogram_plot->SetThreshold(mThreshold);
//	histogram_plot->PostProcessData();
	if (mpData->size() > 0)
	{
		emit dataChanged( createIndex( 0, MlfModelData::EColHardDecision ), createIndex( rowCount(), MlfModelData::EColHardDecision ) );
	}
}

void MlfModel::PriorChanged(double p)
{
	mPrior = p;
	RecomputeNormalizedScore();
}

void MlfModel::MultiplicativeChanged(double m)
{
	mMultiplicative = m;
	RecomputeNormalizedScore();
}

void MlfModel::AdditiveChanged(double a)
{
	mAdditive = a;
	RecomputeNormalizedScore();
}

void MlfModel::RecomputeNormalizedScore()
{
	mComputeScoreThread.Compute(mInputScoreType, mScoreType, mPrior, mMultiplicative, mAdditive, mThreshold);
}

void MlfModel::NormalizedScoreRecomputed(double min, double max)
{
	mNormalizedScoreMin = min;
	mNormalizedScoreMax = max;
	emit dataChanged( createIndex( 0, MlfModelData::EColScoreNorm ), createIndex( rowCount(), MlfModelData::EColScoreNorm ) );
//	ThresholdChanged(mThreshold); // update the hard-decision column
	// draw the histogram
	if (mpHistogramFrame)
	{
		QPainter painter(mpHistogramFrame);
		painter.drawRect(mpHistogramFrame->rect().adjusted(5, 5, -5, -5));
	}
}

void MlfModel::ChangeCorrect(int row)
{
	DBG("MlfModel::ChangeCorrect("<<row<<")");
	MlfModelData::MlfTableRow* r = (*mpData)[row];
	switch(r->mCorrect)
	{
		case MlfModelData::CorrectNo: r->mCorrect = MlfModelData::CorrectUndefined; break;
		case MlfModelData::CorrectYes: r->mCorrect = MlfModelData::CorrectNo; break;
		case MlfModelData::CorrectUndefined: r->mCorrect = MlfModelData::CorrectYes; break;
	}
	emit dataChanged(createIndex(row,MlfModelData::EColCorrect), createIndex(row,MlfModelData::EColCorrect));
}

