#ifndef MLFMODELDATA_H
#define MLFMODELDATA_H

#include <QVector>
#include <QVariant>
#include <QSettings>
#include "mlf.h"
#include "dbg.h"

struct MlfModelData
{
	public:
		enum ECorrect {
			CorrectNo,
			CorrectYes,
			CorrectUndefined
		};

		enum EColumn {
			EColFilename,
			EColStartTime,
			EColEndTime,
			EColLabel,
			EColScoreRaw,
			EColScorePosterior,
			EColScoreLlr,
			EColScoreLlrNorm,
			EColScoreNorm,
			EColHardDecision,
			EColCorrect,
			EColCOUNT
		};

		struct MlfTableRow
		{
			public:
				MlfTableRow(): mpMlfRecord(NULL), mpRefMlfRecord(NULL), mCorrect(CorrectUndefined) {}
				MlfRecord* mpMlfRecord;
				MlfRecord* mpRefMlfRecord;
				ECorrect mCorrect;
				std::string mFilename;
				QString mLabPath;
				QString mLabelUnicode;
				double mScoreLlr;
				double mScoreLlrNorm;
				double mScorePosterior;
				unsigned int mStableSortOrder;

				QVariant operator[] (int i) const
				{
					switch (i) {
						case EColFilename: return mFilename.c_str(); break;
						case EColStartTime: return (qulonglong)mpMlfRecord->mStartTime; break;
						case EColEndTime: return (qulonglong)mpMlfRecord->mEndTime; break;
						case EColLabel: return mLabelUnicode; break;
						case EColScoreRaw: return mpMlfRecord->mScore1; break;
						case EColScorePosterior: return mScorePosterior; break;
						case EColScoreLlr: return mScoreLlr; break;
						case EColScoreLlrNorm: return mScoreLlrNorm; break;
						case EColScoreNorm: return mpMlfRecord->mScore2; break;
						case EColHardDecision: return mpMlfRecord->mHardDecision; break;
						case EColCorrect: return mCorrect; break;
						default: return QVariant(); break;
					}
				}

				friend std::ostream& operator<<(std::ostream& os, const MlfTableRow& r)
				{
					return os << r.mFilename << " " << *(r.mpMlfRecord) << " correct:"<<r.mCorrect<<" scoreLLR:"<<r.mScoreLlr;
				}
		};

		class Columns
		{
			public:
				Columns() : mpSettings(NULL) {}
				// mapping from table column's index to EColumn
				void SetSettings(QSettings * pSettings) { mpSettings = pSettings; }
				typedef std::map<unsigned int, EColumn> Map;
				static QString GetColumnLabel(EColumn col)
				{
					switch (col)
					{
						case EColFilename: return QObject::tr("Filename");
						case EColStartTime: return QObject::tr("Start time");
						case EColEndTime: return QObject::tr("End time");
						case EColLabel: return QObject::tr("Label");
						case EColScoreRaw: return QObject::tr("Raw score");
						case EColScorePosterior: return QObject::tr("Posterior score");
						case EColScoreLlr: return QObject::tr("LLR score");
						case EColScoreLlrNorm: return QObject::tr("Normalized LLR score");
						case EColScoreNorm: return QObject::tr("Normalized 0..1 score");
						case EColHardDecision: return QObject::tr("HD");
						case EColCorrect: return QObject::tr("Correct");
						case EColCOUNT: return QObject::tr("Undef"); /*CERR("EColCOUNT has no label"); EXIT(); break;*/
					}
					return QObject::tr("");
				}
				static QString GetColumnId(EColumn col)
				{
					switch (col)
					{
						case EColFilename: return QObject::tr("Filename");
						case EColStartTime: return QObject::tr("Start_time");
						case EColEndTime: return QObject::tr("End_time");
						case EColLabel: return QObject::tr("Label");
						case EColScoreRaw: return QObject::tr("Raw_score");
						case EColScorePosterior: return QObject::tr("Posterior_score");
						case EColScoreLlr: return QObject::tr("LLR_score");
						case EColScoreLlrNorm: return QObject::tr("Normalized_LLR_score");
						case EColScoreNorm: return QObject::tr("Normalized_0_1_score");
						case EColHardDecision: return QObject::tr("HD");
						case EColCorrect: return QObject::tr("Correct");
						case EColCOUNT: return QObject::tr("Undef"); /*CERR("EColCOUNT has no label"); EXIT(); break;*/
					}
					return QObject::tr("");
				}
				void Clear() 
				{ 
					mMap.clear(); 
					if (mpSettings)
					{
						for (int i=0; i<MlfModelData::EColCOUNT; i++)
						{
							QString key = QString("columns/%1").arg(MlfModelData::Columns::GetColumnId((EColumn)i));
							mpSettings->setValue(key, false);
						}
					}
				}
				void AddColumn(EColumn col) 
				{ 
					mMap.insert(std::make_pair(mMap.size(), col)); 
					if (mpSettings)
					{
						QString key = QString("columns/%1").arg(MlfModelData::Columns::GetColumnId(col));
						mpSettings->setValue(key, true);
					}
				}
				void SetColumn(unsigned int index, EColumn col) { mMap[index] = col; }
				void RemoveColumn(EColumn col) 
				{
					for (Map::iterator it = mMap.begin(); it != mMap.end(); it++)
					{
						if (it->second == col)
						{
							mMap.erase(it);
							return;
						}
					}
				}
				void RemoveIndex(unsigned int i)
				{
					Map::iterator it = mMap.find(i); 
					if (it != mMap.end())
					{
						mMap.erase(it); 
					}
				}
				unsigned int Size() const { return mMap.size(); }
				static unsigned int TotalCount() { return EColCOUNT; }
				int operator[] (unsigned int i) const
				{
					Map::const_iterator it = mMap.find(i);
					if (it != mMap.end())
					{
						return it->second;
					}
					else
					{
						return -1;
					}
				}

			private:
				Map mMap;
				QSettings * mpSettings;
		};

		enum EScoreType {
			LogLikelihoodRatio,
			KarlovcoScore
		};

		enum EInputScoreType {
			InputScoreTypeLogprob, 
			InputScoreTypeLlr
		};

		typedef QVector<MlfTableRow*> Data;
};

#endif

