#include "mlfsortingthread.h"
#include <iostream>
#include <cmath>
using namespace std;

MlfSortingThread::MlfSortingThread(QObject *parent)
    : QThread(parent)
{
    mRestart = false;
    mAbort = false;
	mRunning = false;
	mpMapToSource = new SortingMap;
	mpMapFromSource = new SortingMap;
	mStableSort = false;
	mShowOnlyHD1 = false;
	mFilterChanged = false;
	mSelectedRowsCount = 0;
	mColumnPrev = -1;
	mOrderPrev = -1;
}

MlfSortingThread::~MlfSortingThread()
{
    mMutex.lock();
    mAbort = true;
    mCondition.wakeOne();
    mMutex.unlock();

    wait();
}


void MlfSortingThread::SetColumns(MlfModelData::Columns* pColumns) 
{
	mpColumns = pColumns;
}

void MlfSortingThread::SetData(MlfModelData::Data* pData) 
{ 
	mpData = pData; 
}

void MlfSortingThread::SynchronizeMap()
{
	mMutex.lock();
//	DBG("stop sorting");
	// stop sorting
	mAbort = true;
	mCondition.wakeOne();
	mMutex.unlock();

	DBG("wait");
	wait();
	DBG("wait ...finished");
	mAbort = false;
	mMutex.lock();

	DBG("MlfSortingThread::SynchronizeMap() map_size:"<<mpMapToSource->size()<<" data_size:"<<mpData->size());

	if (mFilterChanged)
	{
		mFilterChanged = false;
	}
	// if the number of rows didn't change, do nothing
/*	else if (mpMapToSource->size() == mpData->size())
	{
		DBG("return");
		mMutex.unlock();
		return;
	}
*/
//	SortingMap mapToSource_old = *mpMapToSource;
//	SortingMap mapFromSource_old = *mpMapFromSource;

	mpMapToSource->clear();
	mpMapFromSource->clear();
	mSelectedRowsCount = 0;

	// add new labels to the selected keywords map
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		MlfModelData::MlfTableRow * row = *i;
		if (mSelectedKeywordsMap.find(row->mLabelUnicode) == mSelectedKeywordsMap.end())
		{
			mSelectedKeywordsMap[row->mLabelUnicode] = true;
		}
	}

	// mapToSource & mapFromSource initialization
	unsigned int filtered_row_id = 0;
	unsigned int row_id = 0;
	for (MlfModelData::Data::iterator i=mpData->begin(); i!=mpData->end(); i++)
	{
		MlfModelData::MlfTableRow * row = *i;
		// 1 1   1
		// 1 0   0
		// 0 1   1
		// 0 0   1
		if (mSelectedKeywordsMap[row->mLabelUnicode] && (mShowOnlyHD1 ? row->mpMlfRecord->mHardDecision : 1))
		{
			mpMapToSource->insert(filtered_row_id, row_id);
			mpMapFromSource->insert(row_id, filtered_row_id);

			++mSelectedRowsCount;
			++filtered_row_id;
		}
		++row_id;
	}

//	DBG("restarting");
	mMutex.unlock();
	if (mRunning)
	{
		start(LowPriority);
	}
	DBG("SynchronizeMap() ...done");
//	Sort(mColumn, mOrder);
}

void MlfSortingThread::FilterChanged()
{
	mFilterChanged = true;
	SynchronizeMap();
}

unsigned int MlfSortingThread::MapToSource(unsigned int i) const
{
	Q_ASSERT(mpMapToSource->find(i) != mpMapToSource->end());
	return (*mpMapToSource)[i];
}

unsigned int MlfSortingThread::MapFromSource(unsigned int i) const
{
	Q_ASSERT(mpMapFromSource->find(i) != mpMapFromSource->end());
	return (*mpMapFromSource)[i];
}

void MlfSortingThread::Sort(int column, Qt::SortOrder order, bool force)
{
	DBG("MlfSortingThread::Sort("<<column<<", "<<order<<", "<<force<<")");
    QMutexLocker locker(&mMutex);

	mInvert = false;
			 	
//	mColumnPrev = mColumn;
	mColumn = (*mpColumns)[column];
//	mOrderPrev = mOrder;
    mOrder = order;
	mForce = force;

    if (!isRunning()) {
        start(LowPriority);
    } else {
        mRestart = true;
        mCondition.wakeOne();
    }
}

void MlfSortingThread::run()
{
    forever 
	{
		DBG("MlfSortingThread start..." );
		mRunning = true;

        mMutex.lock();
		DBG("mutex passed");
		Qt::SortOrder order = mOrder;
		int column = mColumn;
		int column_prev = mColumnPrev;
		bool invert = false;
		MlfModelData::Data* data = mpData;
		SortingMap* map_to_source = mpMapToSource;
		SortingMap* map_from_source = mpMapFromSource;
        mMutex.unlock();


		if (!mForce)
		{
			DBG("MlfSortingThread::run() column:"<<column);
			if (
				column == column_prev
				|| 
				( (column == MlfModelData::EColScoreRaw || 
				   column == MlfModelData::EColScoreLlr || 
				   column == MlfModelData::EColScoreLlrNorm || 
				   column == MlfModelData::EColScorePosterior || 
				   column == MlfModelData::EColScoreNorm ||
				   column == MlfModelData::EColHardDecision)
				  &&
				  (column_prev == MlfModelData::EColScoreRaw || 
				   column_prev == MlfModelData::EColScoreLlr || 
				   column_prev == MlfModelData::EColScoreLlrNorm || 
				   column_prev == MlfModelData::EColScorePosterior || 
				   column_prev == MlfModelData::EColScoreNorm ||
				   column_prev == MlfModelData::EColHardDecision)
				)
			   )
			{
				if (mOrder != mOrderPrev)
				{
					DBG("going to invert the sorting order");
					invert = true;
				}
				else
				{
					emit Sorted();
					return; // everything should be already sorted
				}
			}
		}

		emit SortProgressStart();
		mForce = true;

		if (invert)
		{
			DBG("invert");
			int size = map_to_source->size();
			mProgressPercent = 0;
			for (int i=0; i<size/2; i++)
			{
				if (mRestart)
					break;
				if (mAbort)
					return;

				unsigned int percent = (int)((double)i/(double)size*200);
				if (percent > mProgressPercent)
				{
					emit SortProgress(percent);
					mProgressPercent = percent;
				}

				int i2 = size-i-1;
				int tmp = (*map_to_source)[i];
				(*map_to_source)[i] = (*map_to_source)[i2];
				(*map_to_source)[i2] = tmp;

				tmp = (*map_from_source)[i];
				(*map_from_source)[i] = (*map_from_source)[i2];
				(*map_from_source)[i2] = tmp;
			}
		}
		else
		{
			DBG("!invert");

			bool (*fCmp)(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b) = NULL;

			if (mStableSort)
			{
				switch (static_cast<MlfModelData::EColumn>(column))
				{
					case MlfModelData::EColFilename: fCmp = order==Qt::AscendingOrder ? CmpFilenameAscStable : CmpFilenameDescStable; break;
					case MlfModelData::EColStartTime: fCmp = order==Qt::AscendingOrder ? CmpStartTimeAscStable : CmpStartTimeDescStable; break;
					case MlfModelData::EColEndTime: fCmp = order==Qt::AscendingOrder ? CmpEndTimeAscStable : CmpEndTimeDescStable; break;
					case MlfModelData::EColLabel: fCmp = order==Qt::AscendingOrder ? CmpLabelAscStable : CmpLabelDescStable; break;
					case MlfModelData::EColScoreRaw: fCmp = order==Qt::AscendingOrder ? CmpScoreRawAscStable : CmpScoreRawDescStable; break;
					case MlfModelData::EColScoreLlr: fCmp = order==Qt::AscendingOrder ? CmpScoreLlrAscStable : CmpScoreLlrDescStable; break;
					case MlfModelData::EColScoreLlrNorm: fCmp = order==Qt::AscendingOrder ? CmpScoreLlrNormAscStable : CmpScoreLlrNormDescStable; break;
					case MlfModelData::EColScorePosterior: fCmp = order==Qt::AscendingOrder ? CmpScorePosteriorAscStable : CmpScorePosteriorDescStable; break;
					case MlfModelData::EColScoreNorm: fCmp = order==Qt::AscendingOrder ? CmpScoreNormalizedAscStable : CmpScoreNormalizedDescStable; break;
					case MlfModelData::EColHardDecision: fCmp = order==Qt::AscendingOrder ? CmpScoreNormalizedAscStable : CmpScoreNormalizedDescStable; break;
					case MlfModelData::EColCorrect: fCmp = order==Qt::AscendingOrder ? CmpCorrectAscStable : CmpCorrectDescStable; break;
					case MlfModelData::EColCOUNT: CERR("Sorting by a non-existing column"); EXIT(); break;
				}
			} 
			else 
			{
				switch (static_cast<MlfModelData::EColumn>(column))
				{
					case MlfModelData::EColFilename: fCmp = order==Qt::AscendingOrder ? CmpFilenameAsc : CmpFilenameDesc; break;
					case MlfModelData::EColStartTime: fCmp = order==Qt::AscendingOrder ? CmpStartTimeAsc : CmpStartTimeDesc; break;
					case MlfModelData::EColEndTime: fCmp = order==Qt::AscendingOrder ? CmpEndTimeAsc : CmpEndTimeDesc; break;
					case MlfModelData::EColLabel: fCmp = order==Qt::AscendingOrder ? CmpLabelAsc : CmpLabelDesc; break;
					case MlfModelData::EColScoreRaw: fCmp = order==Qt::AscendingOrder ? CmpScoreRawAsc : CmpScoreRawDesc; break;
					case MlfModelData::EColScoreLlr: fCmp = order==Qt::AscendingOrder ? CmpScoreLlrAsc : CmpScoreLlrDesc; break;
					case MlfModelData::EColScoreLlrNorm: fCmp = order==Qt::AscendingOrder ? CmpScoreLlrNormAsc : CmpScoreLlrNormDesc; break;
					case MlfModelData::EColScorePosterior: fCmp = order==Qt::AscendingOrder ? CmpScorePosteriorAsc : CmpScorePosteriorDesc; break;
					case MlfModelData::EColScoreNorm: fCmp = order==Qt::AscendingOrder ? CmpScoreNormalizedAsc : CmpScoreNormalizedDesc; break;
					case MlfModelData::EColHardDecision: fCmp = order==Qt::AscendingOrder ? CmpScoreNormalizedAsc : CmpScoreNormalizedDesc; break;
					case MlfModelData::EColCorrect: fCmp = order==Qt::AscendingOrder ? CmpCorrectAsc : CmpCorrectDesc; break;
					case MlfModelData::EColCOUNT: CERR("Sorting by a non-existing column"); EXIT(); break;
				}
			}
	
			DBG("fCmp: "<<(int)fCmp);

			if (fCmp)
			{
				//DBG("map_to_source:"<<(int)map_to_source<<" map_from_source:"<<(int)map_from_source<<" data:"<<(int)data);
				//DBG("map_to_source:"<<map_to_source->size()<<" map_from_source:"<<map_from_source->size()<<" data:"<<data->size());
				mProgressCur = 0;
				mProgressPercent = 0;
				mProgressMax = data->size();
				DBG("call sorting function...");
				if (mStableSort)
				{
					//BubbleSort(*data, *map_to_source, *map_from_source, fCmp);
					//MergeSort(*data, *map_to_source, *map_from_source, 0, map_to_source->size()-1, fCmp);
					for (int i=0; i<map_from_source->size(); i++)
					{
						//DBG("mStableSortOrder["<<i<<"]: "<<(*map_from_source)[i]);
						//(*data)[i]->mStableSortOrder = (*map_from_source)[i];

						DBG("mStableSortOrder["<<i<<" -> "<<(*map_to_source)[i]<<"]: "<<i);
						(*data)[(*map_to_source)[i]]->mStableSortOrder = i;
					}
					QuickSort(*data, *map_to_source, *map_from_source, 0, map_to_source->size()-1, fCmp);
				}
				else
				{
					QuickSort(*data, *map_to_source, *map_from_source, 0, map_to_source->size()-1, fCmp);
				}
				DBG("call sorting function...done");
				//SimpleSort(*(mpMlfModel->GetData()), mMapToSource, mMapFromSource, CmpScoreNormalized, order);
			}
		}

		if (mAbort) {
			DBG("mAbort");
			emit Sorted();
			return;
		}

		if (!mRestart)
		{
			mColumnPrev = column;
			mOrderPrev = order;
			mForce = false;
			emit Sorted();
		}

		DBG("MlfSortingThread finished" );

        mMutex.lock();
        if (!mRestart)
		{
			if (mAbort)
			{
				DBG("mAbort");
				return;
			}
            mCondition.wait(&mMutex);
		} else {
			DBG("MlfSortingThread restarted" );
		}
        mRestart = false;
        mMutex.unlock();
	}
}

// SORTING {{{
void MlfSortingThread::Swap(SortingMap& mapToSource, SortingMap& mapFromSource, int i, int j)
{
	unsigned int temp = mapToSource[i];
	mapToSource[i] = mapToSource[j];
	mapToSource[j] = temp;

	temp = mapFromSource[i];
	mapFromSource[i] = mapFromSource[j];
	mapFromSource[j] = temp;
}

void MlfSortingThread::BubbleSort
(
 const MlfModelData::Data& modelData,
 SortingMap& mapToSource,
 SortingMap& mapFromSource,
 bool (*fCmpLess)(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
 )
{
	bool swapped;
	do
	{
		mProgressCur++;
		unsigned int percent = (int)((double)mProgressCur/(double)mProgressMax*100);
		if (percent > mProgressPercent)
		{
			emit SortProgress(percent);
			mProgressPercent = percent;
		}

		swapped = false;
		for (int i=0; i<mapToSource.size()-1; i++)
		{
			if (mRestart || mAbort) 
			{
				return;
			}
			if ((*fCmpLess)(*modelData[mapToSource[i]], *modelData[mapToSource[i+1]]))
			{
				Swap(mapToSource, mapFromSource, i, i+1);
				swapped = true;
			}
		}
	} while(swapped);
}

void MlfSortingThread::QuickSort
(	
 const MlfModelData::Data& modelData,
 SortingMap& mapToSource,
 SortingMap& mapFromSource,
 int first, 
 int last,
 bool (*fCmpLess)(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
 )
{
	mProgressCur++;
//	DBG("mProgressCur:" << mProgressCur << " mProgressMax:"<<mProgressMax<<" mProgressCur/mProgressMax*100="<<(double)mProgressCur/(double)mProgressMax*100);
	unsigned int percent = (int)((double)mProgressCur/(double)mProgressMax*100);
	if (percent > mProgressPercent)
	{
		emit SortProgress(percent);
		mProgressPercent = percent;
	}
	if (mRestart || mAbort) 
	{
		return;
	}
	if(first < last)
	{
		MlfModelData::MlfTableRow* t = modelData[mapToSource[first]];		// t is the pivot. 
		int lastLow = first;
		for (int i = first + 1; i <= last; i++)
		{
			if (mRestart || mAbort) 
			{
				return;
			}
			if((*fCmpLess)(*modelData[mapToSource[i]], *t))
			{
				++lastLow;
				Swap(mapToSource, mapFromSource, lastLow, i);
			}
		}
		Swap(mapToSource, mapFromSource, first, lastLow);
		if(lastLow != first) 
		{
			QuickSort(modelData, mapToSource, mapFromSource, first, lastLow - 1, fCmpLess);
		}
		if(lastLow != last)
		{
			QuickSort(modelData, mapToSource, mapFromSource, lastLow + 1,last, fCmpLess);
		}
	}
}

void MlfSortingThread::MergeSort
(
 const MlfModelData::Data& modelData,
 SortingMap& mapToSource,
 SortingMap& mapFromSource,
 int first, 
 int last,
 bool (*fCmpLess)(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
 )
{
	DBG("MergeSort is buggy!!!");
	EXIT();
	
	int i, j, k;
	unsigned int tmp_to_source, tmp_from_source;
	int n = last-first+1;

	DBG("MergeSort("<<first<<".."<<last<<")");
 
//	for (SortingMap::iterator iMap=mapToSource.begin(); iMap!=mapToSource.end(); iMap++) { DBG("mapToSource: "<<iMap.key()<<" -> "<<iMap.value()); }
	mProgressCur++;
	unsigned int percent = (int)((double)mProgressCur/(double)mProgressMax*100);
	if (percent > mProgressPercent)
	{
		emit SortProgress(percent);
		mProgressPercent = percent;
	}
	if (mRestart || mAbort) 
	{
		return;
	}

	/* For small arrays, insertion sort is much faster */
	if (n < 64) {
		for (i = first+1; i <= last; i++) {
			tmp_to_source = mapToSource[i];
			tmp_from_source = mapFromSource[i];
			for (j = i-1; j >= 0 && (*fCmpLess)(*modelData[mapToSource[tmp_to_source]], *modelData[mapToSource[j]]); j--)
			{
				mapToSource[j+1] = mapToSource[j];
				mapFromSource[j+1] = mapFromSource[j];
			}
			mapToSource[j+1] = tmp_to_source;
			mapFromSource[j+1] = tmp_from_source;
		}

//		for (int i=first; i<=last; i++) { DBG("mapToSource: "<<i<<" -> "<<mapToSource[i]); }
//		cin.get();
		return;
	}
 
	int f = n / 2;		/* f = number of elements in first half */
 
        /* Recursively sort both halves */
//	MergeSort(a, f);
	MergeSort(modelData, mapToSource, mapFromSource, first, first+f-1, fCmpLess);
//	MergeSort(a+f, n-f);
	MergeSort(modelData, mapToSource, mapFromSource, first+f, last, fCmpLess);
 
        /* Merge */
	unsigned int *s_to = new unsigned int[n];	/* temporary array to keep the sorted sequence */
	unsigned int *s_from = new unsigned int[n];	/* temporary array to keep the sorted sequence */
 
	for (i = first, j = first+f, k = 0; i < first+f && j <= last;)
	{
		DBG("i:"<<i<<" j:"<<j<<" map[i]:"<<mapToSource[i]<<" map[j]:"<<mapToSource[j]<<" size:"<<modelData.size());
		if ((*fCmpLess)(*modelData[mapToSource[i]], *modelData[mapToSource[j]]))
		{
			s_to[k] = mapToSource[i];
			s_from[k] = mapFromSource[i];
			i++;
		}
		else
		{
			s_to[k] = mapToSource[j];
			s_from[k] = mapFromSource[j];
			j++;
		}
		k++;
	}
 
	while (i < first+f) 
	{ 
		s_to[k] = mapToSource[i]; 
		s_from[k] = mapFromSource[i];
		k++; 
		i++; 
	}
	while (j <= last)
	{
		s_to[k] = mapToSource[j]; 
		s_from[k] = mapFromSource[j];
		k++; 
		j++; 
	}
 
	for (i = first; i <= last; i++) 
	{
		mapToSource[i] = s_to[i];
		mapFromSource[i] = s_from[i];
	}
//	for (int i=first; i<=last; i++) { DBG("mapToSource: "<<i<<" -> "<<mapToSource[i]); }
//	cin.get();
	delete[] s_to;
	delete[] s_from;
}


bool MlfSortingThread::CmpFilenameAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mFilename > b.mFilename;
}

bool MlfSortingThread::CmpFilenameAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mFilename == b.mFilename ? a.mStableSortOrder < b.mStableSortOrder : a.mFilename > b.mFilename;
}

bool MlfSortingThread::CmpFilenameDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mFilename < b.mFilename;
}

bool MlfSortingThread::CmpFilenameDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mFilename == b.mFilename ? a.mStableSortOrder < b.mStableSortOrder : a.mFilename < b.mFilename;
}

bool MlfSortingThread::CmpStartTimeAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mStartTime > b.mpMlfRecord->mStartTime;
}

bool MlfSortingThread::CmpStartTimeAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mStartTime > b.mpMlfRecord->mStartTime;
}

bool MlfSortingThread::CmpStartTimeDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mStartTime < b.mpMlfRecord->mStartTime;
}

bool MlfSortingThread::CmpStartTimeDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mStartTime < b.mpMlfRecord->mStartTime;
}

bool MlfSortingThread::CmpEndTimeAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mEndTime > b.mpMlfRecord->mEndTime;
}

bool MlfSortingThread::CmpEndTimeAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mEndTime > b.mpMlfRecord->mEndTime;
}

bool MlfSortingThread::CmpEndTimeDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mEndTime < b.mpMlfRecord->mEndTime;
}

bool MlfSortingThread::CmpEndTimeDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mEndTime < b.mpMlfRecord->mEndTime;
}

bool MlfSortingThread::CmpLabelAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) < 0;
}

bool MlfSortingThread::CmpLabelAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) == 0 ? a.mStableSortOrder < b.mStableSortOrder : a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) < 0;
}

bool MlfSortingThread::CmpLabelDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) > 0;
}

bool MlfSortingThread::CmpLabelDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) == 0 ? a.mStableSortOrder < b.mStableSortOrder : a.mLabelUnicode.localeAwareCompare(b.mLabelUnicode) > 0;
}

bool MlfSortingThread::CmpScoreRawAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreRawAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreRawDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreRawDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScorePosteriorAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScorePosteriorAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScorePosteriorDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScorePosteriorDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrNormAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrNormAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 > b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrNormDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreLlrNormDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore1 < b.mpMlfRecord->mScore1;
}

bool MlfSortingThread::CmpScoreNormalizedAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore2 > b.mpMlfRecord->mScore2;
}

bool MlfSortingThread::CmpScoreNormalizedAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore2 > b.mpMlfRecord->mScore2;
}

bool MlfSortingThread::CmpScoreNormalizedDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord->mScore2 < b.mpMlfRecord->mScore2;
}

bool MlfSortingThread::CmpScoreNormalizedDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mpMlfRecord == b.mpMlfRecord ? a.mStableSortOrder < b.mStableSortOrder : a.mpMlfRecord->mScore2 < b.mpMlfRecord->mScore2;
}

bool MlfSortingThread::CmpCorrectAsc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mCorrect < b.mCorrect;
}

bool MlfSortingThread::CmpCorrectAscStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mCorrect == b.mCorrect ? a.mStableSortOrder < b.mStableSortOrder : a.mCorrect < b.mCorrect;
}

bool MlfSortingThread::CmpCorrectDesc(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mCorrect > b.mCorrect;
}

bool MlfSortingThread::CmpCorrectDescStable(const MlfModelData::MlfTableRow &a, const MlfModelData::MlfTableRow &b)
{
	return a.mCorrect == b.mCorrect ? a.mStableSortOrder < b.mStableSortOrder : a.mCorrect > b.mCorrect;
}
// }}} SORTING

